module BlockPDPS

include("Util.jl")
include("Tensor.jl")
include("DTI.jl")
include("Helix.jl")

using Printf
using Distributed
using DelimitedFiles
using JLD2, FileIO
using .DTI
using .Helix

export test_dti_helix,
       test_dti_helix_lsq,
       experiments_refparamsearch,
       experiments_adapt,
       experiments_adapt_accel,
       experiments_adapt_linear

const default_save_prefix="img/nlpdhgm_block/dti_"

const default_params=(
    γ = 1e-9,
    #α = 0.05,
    #τ₀ = 0.5,
    #σ₀ = 19,
    #s̄₀ = 10,
    noise_level = 0.3,
    verbose_iter = 100,
    maxiter = 10000,
    save_results = true,
)

const κ = 0.05

const factor_alts=(
    (s̄₀ = 10, α = 0.5, τ₀ = 0.05, σ₀ = (1-κ)/0.05),
    (s̄₀ = 1, α = 0.005, τ₀ = 1, σ₀ = (1-κ)/1),
)

⬿ = merge

const experiments_refparamsearch=(
    ("non$(p.s̄₀)_tau$(τ₀)", dti_nlpdps_nonblocked, p ⬿ (τ₀ = τ₀, σ₀ = (1-κ)/τ₀))
        for τ₀ in [0.01, 0.05, 0.1, 0.5, 1, 5, 10]
        for p in factor_alts
)

const experiments_adapt=(
    alg for p in factor_alts for alg in (
        ("non$(p.s̄₀)", dti_nlpdps_nonblocked, p),
        ("two$(p.s̄₀)_wy", dti_nlpdps_twoblocked, p ⬿ (wstyle=:proportional_y,)),
        ("vxd$(p.s̄₀)_wy2", dti_nlpdps_voxelblocked_dual, p ⬿ (wstyle=:proportional_y2,)),
        ("vxp$(p.s̄₀)_vartau", dti_nlpdps_voxelblocked_primaldual, p ⬿ (wstyle=:var_τ,))
    )
)

const experiments_adapt_accel=(
    alg for p in factor_alts for alg in (
        alg′ for pa in (
            (accel = :basic, accelfactor = 0.5, α_y = 0.5),
            #(accel = :basic, accelfactor = 0.9, α_y = 0.5),
        ) for alg′ in (
            ("non$(p.s̄₀)_accbasic$(pa.accelfactor)", dti_nlpdps_nonblocked, p ⬿ pa),
            ("two$(p.s̄₀)_accbasic$(pa.accelfactor)_wy", dti_nlpdps_twoblocked, p ⬿ pa ⬿ (wstyle=:proportional_y,)),
            ("vxd$(p.s̄₀)_accbasic$(pa.accelfactor)_wy2", dti_nlpdps_voxelblocked_dual, p ⬿ pa ⬿ (wstyle=:proportional_y2,)),
            ("vxp$(p.s̄₀)_accbasic$(pa.accelfactor)_vartau", dti_nlpdps_voxelblocked_primaldual, p ⬿ pa ⬿ (wstyle=:var_τ,)),
        )
    )
)

const experiments_adapt_linear=(
    alg for p in factor_alts for alg in (
        alg′ for pa in (
            (accel = :linear, accelfactor = 0.5, α_y = 0.5, ρ = 0.1*p.α),
        ) for alg′ in (
            ("non$(p.s̄₀)_linear$(pa.accelfactor)_wy", dti_nlpdps_nonblocked, p ⬿ pa),
            ("two$(p.s̄₀)_linear$(pa.accelfactor)_wy", dti_nlpdps_twoblocked, p ⬿ pa ⬿ (wstyle=:proportional_y,)),
            ("vxd$(p.s̄₀)_linear$(pa.accelfactor)_wy2", dti_nlpdps_voxelblocked_dual, p ⬿ pa ⬿ (wstyle=:proportional_y2,)),
            ("vxp$(p.s̄₀)_linear$(pa.accelfactor)_vartau", dti_nlpdps_voxelblocked_primaldual, p ⬿ pa ⬿ (wstyle=:var_τ,)),
        )
    )
)

function test_dti_helix(;
                        visualise=false,
                        experiments=experiments_adapt,
                        save_prefix=default_save_prefix,
                        kwargs...)

    params = default_params ⬿ kwargs

    # Create visualisation subprocess
    if visualise
        rc = RemoteChannel()
        visproc = @spawn bg_visualise(rc, header)
    else
        rc = nothing
    end

    # Loop over algorithm variants
    for (name, alg, alg_params) in experiments
        println("Running experiment \"$name\"")

        params_this = params ⬿ alg_params

        d, header = test_helix_dtidata(noise_level=params_this.noise_level,
                                       s̄₀=params_this.s̄₀)

        alg′ = now -> alg(d;
                          maybe_verbose = visualise_collect,
                          maybe_verbose_state = (rc, now, 0, nothing),
                          params = params_this)

        # The Julia precompiler is a miserable joke, apparently either not
        # crossing module boundaries (https://discourse.julialang.org/t/--precompile---is-not-resulting-in-precompiled-code/5412/5)
        # or actually not caching things. So it is of no use.
        #
        # To get realiable timings, it is therefore safest to just ignore
        # the time spend in the first iteration and initialising each
        # algorithm. So we start with a negative `start_time` here (`nothing`
        # would be clearner, but maybe the type change from Nothing to
        # Float64 would trigger yet another compilation…), and onlyget the
        # pseudo-start-time in `visualise_collect` below.

        # pc_start = secs_ns()
        # pc_ok = precompile(alg′, (typeof(pc_start),))
        # if pc_ok
        #     println("Precompilation time: $(secs_ns()-pc_start)s")
        # else
        #     error("Precompilation failed")
        # end

        x, yλ, (_, start_time, wasted_time, state) = alg′(-1.0)#secs_ns())

        println("Wasted_time: $(wasted_time)s")

        if params.save_results
            println("Saving ", save_prefix * name, ".(txt, jld2, pdf)")

            perffile = save_prefix * name * ".txt"
            datafile = save_prefix * name * ".jld2"
            pdffile = save_prefix * name * ".pdf"

            open(perffile, "w") do io
                print(io, "# params = ", params_this, "\n")
                writedlm(io, [("iter", "function_value", "time")])
                writedlm(io, unfold_state(state))
            end

            @save datafile x yλ state params d

            Helix.visualise(add_img_mask(x), header;
                            display_now = false,
                            outfile = pdffile)
        end
    end

    if visualise
        # Tell subprocess to finish, and wait
        put!(rc, [])
        wait(visproc)
    end

    return
end

function test_dti_helix_lsq(;
                            save_prefix=default_save_prefix,
                            kwargs...)

    params = default_params ⬿ kwargs

    d, header = test_helix_dtidata(noise_level=params.noise_level,
                                   s̄₀=params.s̄₀)

    Helix.visualise(add_img_mask(dti_lsq(d)), header;
                    outfile = save_prefix * "lsq.pdf")
end

@inline function secs_ns()
    return convert(Float64, time_ns())*1e-9
end

function unfold_state(state)
    res = []
    while state != nothing
        (d, state) = state
        push!(res, d)
    end
    return reverse(res)
end

function add_img_mask(img)
    mask = ones(1, size(img)[2:end]...)
    return cat(mask, img; dims=1)
end

function bg_visualise(rc, header)
    while true
        img=take!(rc)
        # Take only the latest image to visualise
        while isready(rc)
            img=take!(rc)
        end
        # We're done if we were fed an empty image
        if isempty(img)
            break
        end
        Helix.visualise(add_img_mask(img), header; display_now=true)
        # Wait for Preview to reload
        #sleep(1)
        #print(stdout, "\e[s\r\e[30C[V]\e[u")
    end
    return
end

function visualise_collect(iter, d, params, x, yλ, (rc, start_time, wasted_time, state))
    verb = params.verbose_iter!=0 && mod(iter, params.verbose_iter) == 0

    if start_time < 0
        # The Julia precompiler is a miserable joke, apparently not crossing module
        # boundaries, so only start timing after the first iteration.
        start_time = secs_ns()
    end

    if verb || iter ≤ 20 || (iter ≤ 200 && mod(iter, 10) == 0)
        verb_start = secs_ns()
        tm = verb_start - start_time - wasted_time
        v = dti_objective_value(d, params, x)

        # Collect a singly-linked list of state to avoid array resizing while iterating
        state=((iter, v, tm), state)

        if verb
            @printf("%d/%d J=%f\n", iter, params.maxiter, v)
            # Remove old image that the visualisation
            # subprocess has not had the time to process
            if rc != nothing
                while isready(rc)
                    take!(rc)
                end
                put!(rc, x)
            end
        end
        wasted_time += (secs_ns() - verb_start)
    end

    return (rc, start_time, wasted_time, state)
end

end # Module
