
module DTI

using Printf
using ..Tensor
using ..Util

export dti_nlpdps_nonblocked,
       dti_nlpdps_twoblocked,
       dti_nlpdps_voxelblocked_dual,
       dti_nlpdps_voxelblocked_primaldual,
       dti_simple_maybe_verbose,
       dti_objective_value,
       dti_lsq,
       DTIData

struct DTIData
    bvecs :: Array{Array{Float64, 1}, 1}
    s₀ :: Array{Float64, 3}
    s :: Array{Array{Float64, 3}, 1}
end

Primal = Array{Float64,4}
Dual_λ = Array{Float64,4}
Dual_y = Array{Float64,4}
Dual = Tuple{Dual_λ, Dual_y}

const default_params=(
    #α = 0.05,
    #γ = 1e-9,
    ρ = 0,
    τ₀ = 0.05,
    σ₀ = 1.9,
    verbose_iter = 100,
    maxiter = 10000,
    accel = :none,
    accelfactor = 0.5,
    w₀ = 1.0,
    wstyle = :direct,
)

function get_params(params′::NamedTuple)
    params = merge(default_params, params′)
    try
        return params.α, params.γ, params.ρ, params.τ₀, params.σ₀, params
    catch
        throw("Missing parameters")
    end
end


function stejskal_tanner!(λ::Dual_λ, d::DTIData, x::Primal)
    @simd for i in CartesianIndices(d.s₀)
        @inbounds @simd for j in LinearIndices(d.bvecs)
            b = d.bvecs[j]
            λ[j, i] = d.s[j][i]-d.s₀[i]*exp(-symmat₃ₓ₃_vector_iprod(@view(x[:, i]), b))
        end
    end
end

function add_stejskal_tanner_conjgrad_apply!(Δx::Primal, d::DTIData, x::Primal, λ::Dual_λ)
    for j in LinearIndices(d.bvecs)
        b = d.bvecs[j]
        bm = vector_to_symmat₃ₓ₃(b)
        @inbounds @simd for i in CartesianIndices(size(d.s₀))
            # We don't need to weight the calculations here with weights_symmat₃ₓ₃
            # to the correct inner product, as from the expression we are
            # differentiating, exp(-⟨b, mat(x) b⟩)= exp(-⟨b ⊗ b, X⟩) for X=mat(x)
            # the tensor corresponding to the vector presentation x. We can for
            # this explicitly calculate as the gradient exp(-⟨b ⊗ b, X⟩)(b ⊗ b)
            # with respect to the norms and inner products in the space of
            # tensors X.
            #
            # The addition to Δx *needs* to be performed for all but first b-vector,
            # and by the designed use of this function, adds always
            w = d.s₀[i]*exp(-symmat₃ₓ₃_vector_iprod(@view(x[:, i]), b))*λ[j, i]
            # Some weird Julia bug causes this to make insane memory allocs
            #Δx[:, i] .+= w.*bm
            @simd for k=1:length(bm)
                Δx[k, i] .+= w*bm[k]
            end
        end
    end
end

# Our primal objective function is
#
#   J(x) := |T(x)|^2/2 + α|ℰx|_{F,1}.
#
function dti_objective_value(d::DTIData, params::NamedTuple, x::Primal)
    imdim = size(d.s₀)
    y = zeros_symten₃ₓ₃ₓ₃(imdim)
    λ = zeros(length(d.bvecs), imdim...)

    stejskal_tanner!(λ, d, x)
    symmat₃ₓ₃ℰ!(y, x)

    return norm₂²(λ)/2 + params.α*γnorm₂₁(y, params.γ)
end

function dti_simple_maybe_verbose(iter::Int, d::DTIData, params::NamedTuple,
                                  x::Primal, (y, λ)::Dual,
                                  state=nothing)
    if params.verbose_iter!=0 && mod(iter, params.verbose_iter) == 0
        v = dti_objective_value(d, params, x)
        @printf("%d/%d J=%f\n", iter, params.maxiter, v)
        return true
    end
    return false
end

function init_iterates(d::DTIData, xinit::Primal)
    imdim=size(d.s₀)

    x = xinit
    y = zeros_symten₃ₓ₃ₓ₃(imdim)
    λ = zeros(length(d.bvecs), imdim...)
    Δx = similar(x)
    Δy = similar(y)
    Δλ = similar(λ)
    x̄ = similar(x)

    return x, y, λ, Δx, Δy, Δλ, x̄
end

function init_iterates(d::DTIData, xinit::Nothing)
    return init_iterates(d, zeros_symmat₃ₓ₃(size(d.s₀)))
end

function stejskal_tanner_funcs(d::DTIData)
    T! = (λ, x) -> stejskal_tanner!(λ, d, x)
    add_∇T! = (Δx, x, λ) -> add_stejskal_tanner_conjgrad_apply!(Δx, d, x, λ)

    n_b = length(d.bvecs)
    r² = zeros(n_b, size(d.s₀)...)

    for i = 1:n_b
        r²[i, :, :, :] .= (d.s₀.*norm₂²(d.bvecs[i])).^2
    end

    R_T²=sum(r²)

    return T!, add_∇T!, R_T², r²
end

function symgrad_funcs(d::DTIData)
    # h=1.0./size(d.s₀)
    # ℰ! = (y, x) -> symmat₃ₓ₃ℰ!(y, x; h=h)
    # ℰᵀ! = (x, y) -> symmat₃ₓ₃ℰᵀ!(x, y; h=h)
    # R_ℰ²=12/minimum(h)^2
    ℰ! = symmat₃ₓ₃ℰ!
    ℰᵀ! = symmat₃ₓ₃ℰᵀ!
    R_ℰ² = 12

    return ℰ!, ℰᵀ!, R_ℰ²
end

# The saddle point problem corresponding to the primal objective is
#
#   min_x max_{λ,y} ⟨T(x),λ⟩ + ⟨ℰx,y⟩ - G^*(λ) - F^*(y),
#
# where G^*(λ)=|λ|^2/2 and F^*(y)=δ_{αB}(y) for a regularisation parameter
# α and the voxelwise ball 𝔹 on 3×3×3 symmetric tensors.
# Thus the basic algorithm is
#
# x⁺ := x - τ[ℰᵀy + ∇T(x)ᵀλ]
# x̄⁺ := 2x⁺ - x
# y⁺ := proj_{α𝔹}(y + σ[ℰx̄⁺ + T(x̄⁺)]
#
@inline function dti_nlpdps_step!(x, y, λ, Δx, Δy, Δλ, x̄, τ, σ_λ, σ_y,
                                  ω, α, γ, ρ, T!, add_∇T!, ℰ!, ℰᵀ!)
    ℰᵀ!(Δx, y)                        # primal step:
    add_∇T!(Δx, x, λ)                 # |
    @. x̄ = x                          # |  save old x for over-relax
    if ρ>0                            # |  prox
        @. x = (x-τ*Δx)/(1+τ*ρ)       # |  |
    else                              # |  |
        @. x -= τ*Δx                  # |  |
    end                               # |  |
    @. x̄ = (1+ω)*x - ω*x̄              # over-relax: x̄ = 2x-x_old
    T!(Δλ, x̄)                         # dual step: λ
    @. λ = (λ + σ_λ*Δλ)/(1 + σ_λ)     # |  prox
    ℰ!(Δy, x̄)                         # dual step: y
    @. y = (y + σ_y*Δy)/(1 + γ*σ_y/α) # |
    proj_symten₃ₓ₃ₓ₃!(y, α)           # |  prox
end

function dti_nlpdps_nonblocked(d::DTIData;
                               xinit = nothing,
                               maybe_verbose = dti_simple_maybe_verbose,
                               maybe_verbose_state = nothing,
                               params::NamedTuple = default_params)

    α, γ, ρ, τ₀, σ₀, params = get_params(params)
    x, y, λ, Δx, Δy, Δλ, x̄ = init_iterates(d, xinit)
    T!, add_∇T!, R_T², _ = stejskal_tanner_funcs(d)
    ℰ!, ℰᵀ!, R_ℰ² = symgrad_funcs(d)

    R = √(R_T²+R_ℰ²)
    τ = τ₀/R
    σ = σ₀/R
    ω = 1

    if params.accel ≠ :none
        γ̄ = min(γ/α, 1-params.α_y)*params.accelfactor
        if params.accel == :linear
            @assert(ρ>0 && γ̄>0)
            # Solve c in (4.15) with ϵ_j = 1/2
            c = τ*ρ/2
            ω = 1/(1+2c)
        end
    end

    @printf("R=%g, τ=%g, σ=%g, ω=%g [⇒ τ*R=%g, σ*R=%g]\n", R, τ, σ, ω, τ*R, σ*R)

    for iter=1:params.maxiter
        dti_nlpdps_step!(x, y, λ, Δx, Δy, Δλ, x̄, τ, σ, σ,
                         ω, α, γ, ρ, T!, add_∇T!, ℰ!, ℰᵀ!)

        maybe_verbose_state = maybe_verbose(iter, d, params, x, (y, λ), maybe_verbose_state)

        if params.accel == :basic
            σ = σ/(1 + 2*γ̄*σ)
        end
    end

    return x, (y, λ), maybe_verbose_state
end

function dti_nlpdps_twoblocked(d::DTIData;
                               xinit = nothing,
                               maybe_verbose = dti_simple_maybe_verbose,
                               maybe_verbose_state = nothing,
                               params::NamedTuple = default_params)

    α, γ, ρ, τ₀, σ₀, params = get_params(params)
    x, y, λ, Δx, Δy, Δλ, x̄ = init_iterates(d, xinit)
    T!, add_∇T!, R_T², _ = stejskal_tanner_funcs(d)
    ℰ!, ℰᵀ!, R_ℰ² = symgrad_funcs(d)

    # Choose same τ as basic case
    R = √(R_T²+R_ℰ²)
    τ = τ₀/R
    κ = 1 - σ₀*τ₀

    # Choose σ_y and σ_λ relative to blockwise factors
    if params.wstyle == :equal
        # This gives σ_y=σ_λ=σ₀/R similarly to the
        # non-block algorithm
        w = params.w₀*R_ℰ²/R_T²
    elseif params.wstyle == :proportional_y
        # This gives σ_y=σ₀/R_ℰ similarly to how σ=σ₀/R in blockwise
        # analogy to the non-blocked algorithm. The rule for σ_λ
        # is slightly diferent.
        w = params.w₀*√R_ℰ²/(R-√R_ℰ²)
    elseif params.wstyle == :proportional_λ
        # This gives σ_λ=σ₀/R_T similarly to how σ=σ₀/R in blockwise
        # analogy to the non-blocked algorithm. The rule for σ_y
        # is slightly diferent.
        w = params.w₀*(R-√R_T²)/√R_T²
    elseif params.wstyle == :proportional_λ2
        # This also gives σ_λ=σ₀/R_T
        w = params.w₀*R_ℰ²/R_T²
        τ = τ₀*w/((1+w)*√R_T²)
    else
        w = params.w₀
    end
    σ_y = (1-κ)/(τ*(1+1/w)*R_ℰ²)
    σ_λ = (1-κ)/(τ*(1+w)*R_T²)
    ω = 1

    if params.accel ≠ :none
        γ̄_y = γ/α*params.accelfactor
        γ̄_λ = (1-params.α_y)*params.accelfactor
        if params.accel == :linear
            @assert(ρ>0 && γ̄_y>0 && γ̄_λ>0)
            # Solve c in (4.15) with ϵ_j = 1/2
            c = τ*ρ/2
            ω = 1/(1+2c)
        end
    end

    @printf("R=%g, τ=%g, σ_y=%g, σ_λ=%g, w=%g [⇒ τ*R=%g, σ_y*R_ℰ=%g, σ_λ*R_T=%g]\n",
            R, τ, σ_y, σ_λ, w, τ*R, σ_y*√R_ℰ², σ_λ*√R_T²)

    for iter=1:params.maxiter
        dti_nlpdps_step!(x, y, λ, Δx, Δy, Δλ, x̄, τ, σ_λ, σ_y,
                         ω, α, γ, ρ, T!, add_∇T!, ℰ!, ℰᵀ!)

        maybe_verbose_state = maybe_verbose(iter, d, params, x, (y, λ), maybe_verbose_state)

        if params.accel == :basic
            σ_y = σ_y/(1 + 2*γ̄_y*σ_y)
            σ_λ = σ_λ/(1 + 2*γ̄_λ*σ_λ)
        end
    end

    return x, (y, λ), maybe_verbose_state
end

function dti_nlpdps_voxelblocked_dual(d::DTIData;
                                      xinit = nothing,
                                      maybe_verbose = dti_simple_maybe_verbose,
                                      maybe_verbose_state = nothing,
                                      params::NamedTuple = default_params)

    α, γ, ρ, τ₀, σ₀, params = get_params(params)
    x, y, λ, Δx, Δy, Δλ, x̄ = init_iterates(d, xinit)
    T!, add_∇T!, R_T², r² = stejskal_tanner_funcs(d)
    ℰ!, ℰᵀ!, R_ℰ² = symgrad_funcs(d)

    # Choose same τ as basic case
    R = √(R_T²+R_ℰ²)
    τ = τ₀/R
    n_λ = prod(size(λ))
    n_b = length(d.bvecs)

    if params.wstyle == :equal
        w = R_ℰ²./r²
        w_λ = n_b
    elseif params.wstyle == :proportional_y
        # This gives σ_y=σ₀/R_ℰ similarly to how σ=σ₀/R in blockwise
        # analogy to the non-blocked algorithm. The rule for σ_λ
        # is slightly diferent.
        w = repeat([params.w₀*n_λ*√R_ℰ²/(R-√R_ℰ²)], size(λ)...)
        w_λ = n_b
    elseif params.wstyle == :proportional_y2
        # This also gives σ_y=σ₀/R_ℰ similarly to how σ=σ₀/R in blockwise
        # analogy to the non-blocked algorithm. The rule for σ_λ
        # has r-weighting in w (setting instead r ≡ 1 here would give
        # the rule above).
        r = .√r²
        w̄ = sum(r)/(params.w₀*R/√R_ℰ²-1)
        w = w̄./r
        w_λ = n_b
    elseif params.wstyle == :proportional_λ
        # This gives σ_λ=σ₀/r.
        w_λ = n_b
        w = @. params.w₀*R/√r²-w_λ
    elseif params.wstyle == :proportional_λ2
        # This also gives σ_λ=σ₀/r.
        w_λ = repeat(sum(r², dims=1), n_b, 1, 1, 1)./r²
        w = @. params.w₀*R/√r²-w_λ
    elseif params.wstyle == :one
        w = repeat([params.w₀], size(λ)...)
        w_λ = n_b
    end
    # 1 contribution from y block itself, 1./w from λ-blocks
    σ_y = σ₀*R/((1 .+ sum(1.0 ./ w))*R_ℰ²)
    σ_λ = @. σ₀*R/((w_λ + w)*r²)
    ω=1

    if params.accel ≠ :none
        γ̄_y = γ/α*params.accelfactor
        γ̄_λ = (1-params.α_y)*params.accelfactor
        if params.accel == :linear
            @assert(ρ>0 && γ̄_y>0 && γ̄_λ>0)
            # Solve c in (4.15) with ϵ_j = 1/2
            c = τ*ρ/2
            ω = 1/(1+2c)
        end
    end

    @printf("R=%g, τ=%g, σ_y=%g, σ_λ ∈ [%g, %g] [⇒ τ*R=%g, σ_y*R_ℰ=%g, σ_λ*r ∈ [%g, %g]]\n",
            R, τ, σ_y, minimum(σ_λ), maximum(σ_λ),
            τ*R, σ_y*√R_ℰ², minimum(σ_λ.*.√r²), maximum(σ_λ.*.√r²))

    for iter=1:params.maxiter
        dti_nlpdps_step!(x, y, λ, Δx, Δy, Δλ, x̄, τ, σ_λ, σ_y,
                         ω, α, γ, ρ, T!, add_∇T!, ℰ!, ℰᵀ!)

        maybe_verbose_state = maybe_verbose(iter, d, params, x, (y, λ), maybe_verbose_state)

        if params.accel == :basic
            σ_y = σ_y/(1 + 2*γ̄_y*σ_y)
            @. σ_λ = σ_λ/(1 + 2*γ̄_λ*σ_λ)
        end
    end

    return x, (y, λ), maybe_verbose_state
end

function dti_nlpdps_voxelblocked_primaldual(d::DTIData;
                                            xinit = nothing,
                                            maybe_verbose = dti_simple_maybe_verbose,
                                            maybe_verbose_state = nothing,
                                            params::NamedTuple = default_params)

    α, γ, ρ, τ₀, σ₀, params = get_params(params)
    x, y, λ, Δx, Δy, Δλ, x̄ = init_iterates(d, xinit)
    T!, add_∇T!, R_T², r² = stejskal_tanner_funcs(d)
    ℰ!, ℰᵀ!, R_ℰ² = symgrad_funcs(d)

    R = √(R_T²+R_ℰ²)
    κ = 1-τ₀*σ₀
    n_λ = prod(size(λ))
    n_b = length(d.bvecs)
    #r̄² = dropdims(sum(r², dims=1), dims=1)
    r = .√r²
    w_λ = n_b
    #c = 1
    #w = r.*c
    w = r
    if params.wstyle == :var_τ || params.wstyle == :var_τ_mean
        τ = τ₀./dropdims(maximum((1 ./ w .+ w_λ).*r, dims=1), dims=1)
        if params.wstyle == :var_τ_mean
            τ = τ./mean(τ)*(τ₀/R)
        end
        if params.w₀ ≠ 1
            @. τ = params.w₀*τ+(1-params.w₀)*τ₀/R
        end
    elseif params.wstyle == :fix_τ
        τ = repeat([τ₀/R], size(x)[2:end]...)
    end
    σ_y = (1-κ)./(maximum(τ.*(1 .+ dropdims(sum(w, dims=1), dims=1)))*R_ℰ²)
    σ_λ = (1-κ)./(repeat(reshape(τ, 1, size(τ)...), length(d.bvecs), 1, 1, 1).*(w_λ .+ 1 ./ w).*r²)
    ω=1

    if params.accel ≠ :none
        γ̄_y = γ/α*params.accelfactor
        γ̄_λ = (1-params.α_y)*params.accelfactor
        if params.accel == :linear
            @assert(ρ>0 && γ̄_y>0 && γ̄_λ>0)
            # This may be wrong
            c = @. τ*ρ/2
            ω = @. 1/(1+2c)
        end
    end

    @printf("R=%g, τ ∈ [%g, %g], σ_y=%g, σ_λ ∈ [%g, %g] [⇒ τ*R ∈ [%g, %g], σ_y*R_ℰ=%g, σ_λ*r ∈ [%g, %g]]\n",
            R, minimum(τ), maximum(τ), σ_y, minimum(σ_λ), maximum(σ_λ),
            minimum(τ.*R), maximum(τ.*R), σ_y*√R_ℰ², minimum(σ_λ.*.√r²), maximum(σ_λ.*.√r²))

    for iter=1:params.maxiter
        ℰᵀ!(Δx, y)                        # primal step:
        add_∇T!(Δx, x, λ)                 # |
        @. x̄ = x                          # |  save old x for over-relax
        if ρ>0                            # |  prox & over-relax
            tensor_scalar_mult_subtract_scale!(x, τ, Δx, 1 .+ ρ*τ)
            tensor_overrelax!(x̄, x, ω)    # |  |
        else                              # |  |
            tensor_scalar_mult_subtract!(x, τ, Δx)
            @. x̄ = (1+ω)*x - ω*x̄          # |  |
        end                               # |  |
        T!(Δλ, x̄)                         # dual step: λ
        @. λ = (λ + σ_λ*Δλ)/(1 + σ_λ)     # |  prox
        ℰ!(Δy, x̄)                         # dual step: y
        @. y = (y + σ_y*Δy)/(1 + γ*σ_y/α) # |
        proj_symten₃ₓ₃ₓ₃!(y, α)           # |  prox

        maybe_verbose_state = maybe_verbose(iter, d, params, x, (y, λ), maybe_verbose_state)

        if params.accel == :basic
            σ_y = σ_y/(1 + 2*γ̄_y*σ_y)
            @. σ_λ = σ_λ/(1 + 2*γ̄_λ*σ_λ)
        end
    end

    return x, (y, λ), maybe_verbose_state
end

function dti_lsq(d::DTIData)

    x = zeros_symmat₃ₓ₃(size(d.s₀))

    @simd for i in CartesianIndices(d.s₀)
        a = zeros(length(d.bvecs))
        M = zeros(length(d.bvecs), 6)
        for j in LinearIndices(d.bvecs)
            a[j] = log(max(eps(), d.s[j][i]))-log(d.s₀[i])
            M[j, :] = -weights_symmat₃ₓ₃.*vector_to_symmat₃ₓ₃(d.bvecs[j])
        end
        x[:, i] = (M'*M) \ (M'*a)
    end

    return x
end

end # Module


