
module Helix

using JLD2, FileIO
using NRRD
using Printf
using ..Tensor
using ..Util
using ..DTI

### Workaround to NRRD bug ###

import NRRD.alloctype
import NRRD.PTuple

alloctype(::Type{PTuple{T}}) where {T} = Any # Union{String,Vector{T}}

### end workaround to NRRD bug ###

export test_helix_dtidata,
       test_helix,
       make_test_helix,
       visualise

const test_helix_filename="data/test_helix.nrrd"
const tmp_filename="data/_tmp_visualisation1.nrrd"
const emap_filename="data/_tmp_visualisation2.nrrd"
const tmp_pdf_filename="data/_visualisation.pdf"
const pdf_filename="data/visualisation.pdf"
const slice_pdf_filename="data/slice_visualisation.pdf"
const noisy_data_prefix="data/helix_noisy"
const emap="1 1 1 1 -1 -1 -3"

const bvecs=[
    [1; 0; 0],
    [0; 1; 0],
    [0; 0; 1],
    [√2; √2; 0],
    [√2; 0; √2],
    [0; √2; √2],
    ]

function make_test_helix()
    if !isfile(test_helix_filename)
        run(`tend helix -s 38 39 40 -b 0 -r 0.5 -R 1.2 -S 2 -ev 0.25 0.1 0.05 -bg 0.01 -o $test_helix_filename`)
    end
end


function test_helix()
    make_test_helix()
    header = NRRD.parse_header(test_helix_filename)
    img = load(File(format"NRRD",test_helix_filename))
    return convert(Array{Float64}, img), header
end

function test_helix_dtidata(;noise_level=0.05, s̄₀=10)
    saved_noisy = @sprintf("%s_%g_%g.jld2", noisy_data_prefix, noise_level, s̄₀)

    if isfile(saved_noisy)
        println("Loading $saved_noisy")
        @load saved_noisy d header
    else
        img, header = test_helix()

        # img[1, :, :, :] is the mask, which needs to be dropped
        s₀ = s̄₀*dropdims(sqrt.(sum(img[2:end, :, :, :].^2, dims=1)), dims=1);
        noise_level′ = noise_level*sum(abs.(s₀))/prod(size(s₀));

        make_signal = b -> begin
            sj = d -> begin
                return exp(-symmat₃ₓ₃_vector_iprod(d[1].*d[2:end], b))
            end
            noise=noise_level′*randn(Float64, size(s₀)...)
            return noise+s₀.*dropdims(mapslices(sj, img, dims=1), dims=1);
        end

        s = [make_signal(bvecs[i]) for i=1:length(bvecs)]

        d = DTIData(bvecs, s₀, s)

        println("Saving $saved_noisy")
        @save saved_noisy d header
    end

    return d, header
end

# Need to use a saved header due to limitations of NRRD.jl.
function visualise(img, saved_header; display_now=false, outfile=pdf_filename)
    open(tmp_filename, "w") do io
        write(io, magic(format"NRRD"))
        NRRD.write_header(io, saved_header...)
        NRRD.nrrd_write(io, convert(Array{Float32}, img))
    end
    camera=["-fr", "10", "0", "2",
            "-at", "5", "0", "1",
            "-up", "0", "0", "1", "-rh"]
    run(pipeline(`echo "1 1 1 1 -1 -1 -3"`, `emap -amb 0.5 0.5 0.5 -i - $camera -o $emap_filename`))
    run(pipeline(`tend glyph -i $tmp_filename -emap $emap_filename -bg 1 1 1 -sat 0.8 $camera
                  -ur -1 1 -vr -1 1 -psc 345 -gsc 0.3 -a eval0 -atr 0.1 -ctr 0.8`,
                 `epstopdf --filter -o=$tmp_pdf_filename`))
    # Can't just Base.Filesystem.mv as Preview seems to open a new window if the
    # file inode changes.
    run(pipeline(`cat $tmp_pdf_filename`, "$outfile"))
    if display_now
        run(`open -g $outfile`)
    end
    # run(pipeline(`tend slice -i $tmp_filename -a 2 -p 10 -d`,
    #              `tend glyph -emap $emap_filename -bg 1 1 1 -sat 0.8 $camera
    #               -ur -1 1 -vr 1 1 -wd 1.6 0.8 0 -psc 900 -gsc 0.03 -atr 0.6`,
    #              `epstopdf --filter -o=$slice_pdf_filename`))
    return
end

end # Module



