#
# Some generic functions on specific symmetric tensors
#

module Tensor

using ..Util

export proj_weighted!,
       proj_symten₃ₓ₃ₓ₃!,
       symmat₃ₓ₃ℰ!,
       symmat₃ₓ₃ℰᵀ!,
       zeros_symmat₃ₓ₃,
       zeros_symten₃ₓ₃ₓ₃,
       symmat₃ₓ₃_vector_iprod,
       symmat₃ₓ₃_vector_appl,
       vector_to_symmat₃ₓ₃,
       weights_symmat₃ₓ₃,
       weights_symten₃ₓ₃ₓ₃,
       tensor_scalar_mult_subtract!,
       tensor_scalar_mult_subtract_scale!,
       tensor_overrelax!

# 3×3 symmetric matrix presentation indexing and component weighting

const i_xx = 1
const i_xy, i_yx = 2, 2
const i_xz, i_zx = 3, 3
const i_yy = 4
const i_yz, i_zy = 5, 5
const i_zz = 6

const weights_symmat₃ₓ₃=[
    1, # xx
    2, # xy = yx
    2, # xz = zx
    1, # yy
    2, # yz = zy
    1, # zz
]

# 3×3×3 symmetric tensor presentation indexing and component weighting

const i_xxx = 1
const i_xxy, i_xyx, i_yxx = 2, 2, 2
const i_xxz, i_xzx, i_zxx = 3, 3, 3
const i_xyy, i_yxy, i_yyx = 4, 4, 4
const i_xyz, i_xzy, i_zxy, i_zyx, i_yxz, i_yzx = 5, 5, 5, 5, 5, 5
const i_xzz, i_zxz, i_zzx = 6, 6, 6
const i_yyy = 7
const i_yyz, i_yzy, i_zyy = 8, 8, 8
const i_yzz, i_zyz, i_zzy = 9, 9, 9
const i_zzz = 10

const weights_symten₃ₓ₃ₓ₃=[
    1, # xxx
    3, # xxy = xyx = yxx
    3, # xxz = xzx = zxx
    3, # xyy = yxy = yyx
    6, # xyz = xzy = zxy = zyx = yxz = yzx
    3, # xzz = zxz = zzx
    1, # yyy
    3, # yyz = yzy = zyy
    3, # yzz = zyz = zzy
    1, # zzz
]


function zeros_symmat₃ₓ₃(dims)
    @assert(length(dims)==3)
    return zeros(length(weights_symmat₃ₓ₃), dims...)
end

function zeros_symten₃ₓ₃ₓ₃(dims)
    @assert(length(dims)==3)
    return zeros(length(weights_symten₃ₓ₃ₓ₃), dims...)
end

@inline function proj_weighted!(y, α, w)
    α²=α*α
    y′=reshape(y, (size(y, 1), prod(size(y)[2:end])))

    @inbounds @simd for i=1:size(y′, 2)# in CartesianIndices(size(y)[2:end])
        n² = norm₂w²(@view(y′[:, i]), w)
        if n²>α²
            y′[:, i] .*= (α/√n²)
        end
    end
end

function proj_symten₃ₓ₃ₓ₃!(y, α)
    proj_weighted!(y, α, weights_symten₃ₓ₃ₓ₃)
end

# y = y - τ*x voxelwise
@inline function tensor_scalar_mult_subtract!(y, τ, x)
    l=prod(size(τ))
    τ′=reshape(τ, l)
    y′=reshape(y, (size(y, 1), l))
    x′=reshape(x, (size(x, 1), l))

    @inbounds @simd for i=1:l
        y′[:, i] .-= τ′[i]*x′[:, i]
    end
end

# y = (y - τ*x)/β voxelwise
@inline function tensor_scalar_mult_subtract_scale!(y, τ, x, β)
    l=prod(size(τ))
    τ′=reshape(τ, l)
    β′=reshape(β, l)
    y′=reshape(y, (size(y, 1), l))
    x′=reshape(x, (size(x, 1), l))

    @inbounds @simd for i=1:l
        y′[:, i] .= (y′[:, i] .- τ′[i]*x′[:, i])/β′[i]
    end
end

# x̄ = (1+ω)*x - ω*x̄ voxelwise
@inline function tensor_overrelax!(x̄, x, ω)
    l=prod(size(ω))
    ω′=reshape(ω, l)
    x′=reshape(x, (size(x, 1), l))
    x̄′=reshape(x̄, (size(x̄, 1), l))

    @inbounds @simd for i=1:l
        x̄′[:, i] .= (1+ω′[i])*x′[:, i] - ω′[i]*x̄′[:, i]
    end
end

function symmat₃ₓ₃ℰ!(y,x; h = (1,1,1))
    dx! = (i, j, w₁) -> @inbounds @views @. begin
        y[i, 1:(end-1), :, :] = (x[j, 2:end, :, :] - x[j, 1:(end-1), :, :])*w₁
        y[i, end, :, :] = 0
    end

    dy! = (i, j, w₂) -> @inbounds @views @. begin
        y[i, :, 1:(end-1), :] = (x[j, :, 2:end, :] - x[j, :, 1:(end-1), :])*w₂
        y[i, :, end, :] = 0
    end

    dz! = (i, j, w₃) -> @inbounds @views @. begin
        y[i, :, :, 1:(end-1)] = (x[j, :, :, 2:end] - x[j, :, :, 1:(end-1)])*w₃
        y[i, :, :, end] = 0
    end

    dx_dy! = (i, (j₁,j₂), (w₁,w₂)) -> @inbounds @views @. begin
        y[i, 1:(end-1), 1:(end-1), :] = (
             (x[j₁, 2:end, 1:(end-1), :] - x[j₁, 1:(end-1), 1:(end-1), :])*w₁
            +(x[j₂, 1:(end-1), 2:end, :] - x[j₂, 1:(end-1), 1:(end-1), :])*w₂)
        y[i, 1:(end-1), end, :] = (x[j₁, 2:end, end, :] - x[j₁, 1:(end-1), end, :])*w₁
        y[i, end, 1:(end-1), :] = (x[j₂, end, 2:end, :] - x[j₂, end, 1:(end-1), :])*w₂
        y[i, end, end, :] = 0
    end

    dx_dz! = (i, (j₁,j₂), (w₁,w₃)) -> @inbounds @views @. begin
        y[i, 1:(end-1), :, 1:(end-1)] = (
             (x[j₁, 2:end, :, 1:(end-1)] - x[j₁, 1:(end-1), :, 1:(end-1)])*w₁
            +(x[j₂, 1:(end-1), :, 2:end] - x[j₂, 1:(end-1), :, 1:(end-1)])*w₃)
        y[i, 1:(end-1), :, end] = (x[j₁, 2:end, :, end] - x[j₁, 1:(end-1), :, end])*w₁
        y[i, end, :, 1:(end-1)] = (x[j₂, end, :, 2:end] - x[j₂, end, :, 1:(end-1)])*w₃
        y[i, end, :, end] = 0
    end

    dy_dz! = (i, (j₁,j₂), (w₂,w₃)) -> @inbounds @views @. begin
        y[i, :, 1:(end-1), 1:(end-1)] = (
             (x[j₁, :, 2:end, 1:(end-1)] - x[j₁, :, 1:(end-1), 1:(end-1)])*w₂
            +(x[j₂, :, 1:(end-1), 2:end] - x[j₂, :, 1:(end-1), 1:(end-1)])*w₃)
        y[i, :, 1:(end-1), end] = (x[j₁, :, 2:end, end] - x[j₁, :, 1:(end-1), end])*w₂
        y[i, :, end, 1:(end-1)] = (x[j₂, :, end, 2:end] - x[j₂, :, end, 1:(end-1)])*w₃
        y[i, :, end, end] = 0
    end

    dx_dy_dz! = (i, (j₁,j₂,j₃), (w₁,w₂,w₃)) -> begin
        @inbounds @views @. begin
            y[i, 1:(end-1), 1:(end-1), 1:(end-1)] = (
                 (x[j₁, 2:end, 1:(end-1), 1:(end-1)] - x[j₁, 1:(end-1), 1:(end-1), 1:(end-1)])*w₁
                +(x[j₂, 1:(end-1), 2:end, 1:(end-1)] - x[j₂, 1:(end-1), 1:(end-1), 1:(end-1)])*w₂
                +(x[j₃, 1:(end-1), 1:(end-1), 2:end] - x[j₃, 1:(end-1), 1:(end-1), 1:(end-1)])*w₃)

            y[i, 1:(end-1), 1:(end-1), end] = (
                 (x[j₁, 2:end, 1:(end-1), end] - x[j₁, 1:(end-1), 1:(end-1), end])*w₁
                +(x[j₂, 1:(end-1), 2:end, end] - x[j₂, 1:(end-1), 1:(end-1), end])*w₂)
            y[i, 1:(end-1), end, 1:(end-1)] = (
                 (x[j₁, 2:end, end, 1:(end-1)] - x[j₁, 1:(end-1), end, 1:(end-1)])*w₁
                +(x[j₃, 1:(end-1), end, 2:end] - x[j₃, 1:(end-1), end, 1:(end-1)])*w₃)
            y[i, end, 1:(end-1), 1:(end-1)] = (
                 (x[j₂, end, 2:end, 1:(end-1)] - x[j₂, end, 1:(end-1), 1:(end-1)])*w₂
                +(x[j₃, end, 1:(end-1), 2:end] - x[j₃, end, 1:(end-1), 1:(end-1)])*w₃)

            y[i, 1:(end-1), end, end] = (x[j₁, 2:end, end, end] - x[j₁, 1:(end-1), end, end])*w₁
            y[i, end, 1:(end-1), end] = (x[j₂, end, 2:end, end] - x[j₂, end, 1:(end-1), end])*w₂
            y[i, end, end, 1:(end-1)] = (x[j₃, end, end, 2:end] - x[j₃, end, end, 1:(end-1)])*w₃

        end
        # Julia chokes up if this is inside @views
        y[i, end, end, end] = 0.0
    end

    dx!(i_xxx, i_xx, 1/h[1])
    dy!(i_yyy, i_yy, 1/h[2])
    dz!(i_zzz, i_zz, 1/h[3])
    dx_dy!(i_xxy, (i_xy, i_xx), (2/3, 1/3)./(h[1],h[2]))
    dx_dz!(i_xxz, (i_zx, i_xx), (2/3, 1/3)./(h[1],h[3]))
    dx_dy!(i_xyy, (i_yy, i_xy), (1/3, 2/3)./(h[1],h[2]))
    dx_dz!(i_xzz, (i_zz, i_xz), (1/3, 2/3)./(h[1],h[3]))
    dy_dz!(i_yyz, (i_yz, i_yy), (2/3, 1/3)./(h[2],h[3]))
    dy_dz!(i_yzz, (i_zz, i_yz), (1/3, 2/3)./(h[2],h[3]))
    dx_dy_dz!(i_xyz, (i_yz, i_zx, i_xy), (1/3, 1/3, 1/3)./h)

    return
end

function symmat₃ₓ₃ℰᵀ!(x, y; h = (1,1,1))
    dx_dy_dzᵀ! = (j, (i₁,i₂,i₃), (d₁,d₂,d₃)) -> @inbounds begin
        # Besides averaging in the calculation of the symmetrised gradient ℰ,
        # we need to include the weighting of the inner products in the spaces
        # (that accounts for only storing non-redundant components) for the
        # calculation of the adjoint.
        # For all z that are vector presentations of a symmetric 3×3×3 tensor,
        # and x that are vector presentation of a symmetric 3×3 matrix,
        # with respect to the vector-presentation gradient ℰ, for ⟨⋅,⋅⟩_wm and ⟨⋅,⋅⟩_wt
        # the weighted inner producuts corresponding to the matrix and tensor inner
        # product, for ⊙ the coordinate-wise multiplication set, we have
        #
        #  ⟨ℰx,z⟩_wᵗ = ⟨ℰx, wᵗ ⊙ z⟩_{ℝⁿ} = ⟨x,ℰᵀ [wᵗ ⊙ z]⟩_{ℝᵐ}
        #            = ⟨x,(wᵐ)⁻¹ ⊙ ℰᵀ [wᵗ ⊙ z]⟩_wᵐ
        #
        #  Therefore
        #
        #   ℰ^* = (wᵐ)⁻¹ ⊙ ℰᵀ ⊙ wᵗ
        #
        w₁=(weights_symten₃ₓ₃ₓ₃[i₁]/weights_symmat₃ₓ₃[j])*d₁
        w₂=(weights_symten₃ₓ₃ₓ₃[i₂]/weights_symmat₃ₓ₃[j])*d₂
        w₃=(weights_symten₃ₓ₃ₓ₃[i₃]/weights_symmat₃ₓ₃[j])*d₃

        @views @. begin
            x[j, 2:(end-1), :, :] = (y[i₁, 1:(end-2), :, :] - y[i₁, 2:(end-1), :, :]).*w₁
            x[j, 1, :, :] = -y[i₁, 1, :, :].*w₁
            x[j, end, :, :] = y[i₁, end-1, :, :].*w₁

            x[j, :, 2:(end-1), :] += (y[i₂, :, 1:(end-2), :] - y[i₂, :, 2:(end-1), :]).*w₂
            x[j, :, 1, :] += -y[i₂, :, 1, :].*w₂
            x[j, :, end, :] += y[i₂, :, end-1, :].*w₂

            x[j, :, :, 2:(end-1)] += (y[i₃, :, :, 1:(end-2)] - y[i₃, :, :, 2:(end-1)]).*w₃
            x[j, :, :, 1] += -y[i₃, :, :, 1].*w₃
            x[j, :, :, end] += y[i₃, :, :, end-1].*w₃
        end
    end

    dx_dy_dzᵀ!(i_xx, (i_xxx, i_xxy, i_xxz), (1,   1/3, 1/3)./h)
    dx_dy_dzᵀ!(i_yy, (i_yyx, i_yyy, i_yyz), (1/3, 1,   1/3)./h)
    dx_dy_dzᵀ!(i_zz, (i_zzx, i_zzy, i_zzz), (1/3, 1/3, 1)./h)
    dx_dy_dzᵀ!(i_xy, (i_xyx, i_xyy, i_xyz), (2/3, 2/3, 1/3)./h)
    dx_dy_dzᵀ!(i_xz, (i_xzx, i_xzy, i_xzz), (2/3, 1/3, 2/3)./h)
    dx_dy_dzᵀ!(i_yz, (i_yzx, i_yzy, i_yzz), (1/3, 2/3, 2/3)./h)

    return
end

# Calculate ⟨b,mat(x)b⟩
@inline function symmat₃ₓ₃_vector_iprod(x, b)
    return @inbounds (
             b[1]*(x[i_xx]*b[1]+x[i_xy]*b[2]+x[i_xz]*b[3])
            +b[2]*(x[i_yx]*b[1]+x[i_yy]*b[2]+x[i_yz]*b[3])
            +b[3]*(x[i_zx]*b[1]+x[i_zy]*b[2]+x[i_zz]*b[3]))
end

@inline function symmat₃ₓ₃_vector_appl(x, b)
    return @inbounds [
            x[i_xx]*b[1]+x[i_xy]*b[2]+x[i_xz]*b[3]
            x[i_yx]*b[1]+x[i_yy]*b[2]+x[i_yz]*b[3]
            x[i_zx]*b[1]+x[i_zy]*b[2]+x[i_zz]*b[3]]
end

# Calculate vector presentation of the 3×3 symmetric matrix b ⨵ b
@inline function vector_to_symmat₃ₓ₃(b)
    return @inbounds [
            b[1]*b[1], b[1]*b[2], b[1]*b[3],
                       b[2]*b[2], b[2]*b[3],
                                  b[3]*b[3]]
end

end # Module

